# lingo - payment and billing system
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import datetime

from django.test.client import RequestFactory
from django.utils.translation import ugettext_lazy as _

from lingo.agendas.chrono import get_check_status, get_subscriptions
from lingo.agendas.models import Agenda
from lingo.invoicing.models import DraftInvoice, DraftInvoiceLine, RegieNotConfigured
from lingo.pricing.models import AgendaPricing, AgendaPricingNotFound, PricingError


def get_agendas(pool):
    agendas_pricings = AgendaPricing.objects.filter(flat_fee_schedule=False).extra(
        where=["(date_start, date_end) OVERLAPS (%s, %s)"],
        params=[pool.campaign.date_start, pool.campaign.date_end],
    )
    return Agenda.objects.filter(pk__in=agendas_pricings.values('agendas')).order_by('pk')


def get_users_from_subscriptions(agendas, pool):
    user_external_ids = set()
    for agenda in agendas:
        subscriptions = get_subscriptions(
            agenda_slug=agenda.slug,
            date_start=pool.campaign.date_start,
            date_end=pool.campaign.date_end,
        )
        for subscription in subscriptions:
            user_external_ids.add(subscription['user_external_id'])
    return user_external_ids


def get_invoice_lines_for_user(agendas, agendas_pricings, user_external_id, pool):
    def get_agenda_pricing(agendas_pricings_for_agenda, date_event):
        # same logic as AgendaPricing.get_agenda_pricing
        for agenda_pricing in agendas_pricings_for_agenda:
            if agenda_pricing.date_start > date_event:
                continue
            if agenda_pricing.date_end <= date_event:
                continue
            return agenda_pricing
        raise AgendaPricingNotFound

    if not agendas:
        return []

    agendas_by_slug = {a.slug: a for a in agendas}
    agendas_pricings_by_agendas = collections.defaultdict(list)
    for agenda_pricing in agendas_pricings:
        if agenda_pricing.flat_fee_schedule:
            continue
        for agenda in agenda_pricing.agendas.all():
            agendas_pricings_by_agendas[agenda.slug].append(agenda_pricing)

    # get check status for user_external_id, on agendas, for the period
    check_status_list = get_check_status(
        agenda_slugs=list(agendas_by_slug.keys()),
        user_external_id=user_external_id,
        date_start=pool.campaign.date_start,
        date_end=pool.campaign.date_end,
    )
    request = RequestFactory().get('/')  # XXX

    # build lines from check status
    lines = []
    for check_status in check_status_list:
        serialized_event = check_status['event']
        event_date = datetime.datetime.fromisoformat(serialized_event['start_datetime']).date()
        event_slug = '%s@%s' % (serialized_event['agenda'], serialized_event['slug'])

        agenda = agendas_by_slug[serialized_event['agenda']]
        try:
            agenda_pricing = get_agenda_pricing(agendas_pricings_by_agendas.get(agenda.slug), event_date)
            pricing_data = agenda_pricing.get_pricing_data_for_event(
                request=request,
                agenda=agenda,
                event=serialized_event,
                check_status=check_status['check_status'],
                user_external_id=user_external_id,
                adult_external_id=user_external_id,  # XXX
            )
        except PricingError as e:
            # AgendaPricingNotFound: can happen if pricing model defined only on a part of the requested period
            pricing_error = {
                'error': type(e).__name__,
                'error_details': e.details,
            }
            lines.append(
                DraftInvoiceLine(
                    slug=event_slug,
                    label=serialized_event['label'],
                    quantity=0,
                    unit_amount=0,
                    total_amount=0,
                    user_external_id=user_external_id,
                    payer_external_id=user_external_id,  # XXX
                    event=serialized_event,
                    pricing_data=pricing_error,
                    status='warning' if isinstance(e, AgendaPricingNotFound) else 'error',
                    pool=pool,
                )
            )
        else:
            # XXX log all context !
            lines.append(
                DraftInvoiceLine(
                    slug=event_slug,
                    label=serialized_event['label'],
                    quantity=1,
                    unit_amount=pricing_data['pricing'],
                    total_amount=pricing_data['pricing'],
                    user_external_id=user_external_id,
                    payer_external_id=user_external_id,  # XXX
                    event=serialized_event,
                    pricing_data=pricing_data,
                    status='success',
                    pool=pool,
                )
            )
    DraftInvoiceLine.objects.bulk_create(lines)
    return lines


def get_all_invoice_lines(agendas, user_external_ids, pool):
    agendas_pricings = (
        AgendaPricing.objects.filter(flat_fee_schedule=False)
        .extra(
            where=["(date_start, date_end) OVERLAPS (%s, %s)"],
            params=[pool.campaign.date_start, pool.campaign.date_end],
        )
        .prefetch_related('agendas', 'pricing__criterias', 'pricing__categories')
    )

    lines = []
    for user_external_id in user_external_ids:
        # generate lines for each user
        lines += get_invoice_lines_for_user(
            agendas=agendas,
            agendas_pricings=agendas_pricings,
            user_external_id=user_external_id,
            pool=pool,
        )
    return lines


def generate_invoices_from_lines(agendas, all_lines, pool):
    agendas_by_slug = {a.slug: a for a in agendas}

    # regroup lines by regie, and by payer_external_id (payer)
    lines_by_regie = {}
    for line in all_lines:
        if line.status != 'success':
            # ignore lines in error
            continue
        agenda_slug = line.event['agenda']
        if agenda_slug not in agendas_by_slug:
            # should not happen
            continue
        regie = agendas_by_slug[agenda_slug].regie
        if not regie:
            raise RegieNotConfigured(_('Regie not configured on %s') % agenda_slug)
        if regie.pk not in lines_by_regie:
            lines_by_regie[regie.pk] = {}
        regie_subs = lines_by_regie[regie.pk]
        if line.payer_external_id not in regie_subs:
            regie_subs[line.payer_external_id] = []
        regie_subs[line.payer_external_id].append(line)

    # generate invoices by regie and by payer_external_id (payer)
    invoices = []
    for regie_id, regie_subs in lines_by_regie.items():
        for payer_external_id, adult_lines in regie_subs.items():
            invoice = DraftInvoice.objects.create(
                label=_('Invoice from %(start)s to %(end)s')
                % {
                    'start': pool.campaign.date_start,
                    'end': pool.campaign.date_end - datetime.timedelta(days=1),
                },
                date_issue=pool.campaign.date_issue,
                regie_id=regie_id,
                payer=payer_external_id,
                pool=pool,
            )
            DraftInvoiceLine.objects.filter(pk__in=[line.pk for line in adult_lines]).update(invoice=invoice)
            invoices.append(invoice)

    return invoices
